#include "bonus_points.hpp"

#include "util/enums.hpp"
#include "util/exception.hpp"
#include "util/heuristics.hpp"
#include "util/points.hpp"
#include "util/strings.hpp"

BonusPoints::BonusPoints(const CsvDump &c) noexcept(false) : mBonusPoints(c) {
    if (c.data.empty()) {
        // throw BonusParseException(str::error::empty);
        // empty bonus point file is ok, just no bonus points for anybody
        // todo: why was this not ok in the past?
        return;
    }

    if (c.data[0].empty())
        throw BonusParseException(str::error::noHeader);

    // Quick check: reading wrong file type?
    {
        const auto fileType = getApproximateFileType(c.data[0][0]);

        switch (fileType) {
        case FileType::dynexite:
            throw BonusParseException(str::error::unexpectedDynexite);
        case FileType::rwth:
            throw BonusParseException(str::error::unexpectedRwthOnline);
        default:
            // the file might still be invalid, but that will be decided later
            break;
        }
    }

    const auto nrRows = c.data.size();

    // No data aside from header is allowed
    if (nrRows < 2)
        return;

    mRegistrationNr.reserve(nrRows - 1);
    mNrBonusPoints.reserve(nrRows - 1);

    // Get column index
    const auto registrationNrColumn = c.data[0].indexOf("Matrikelnummer");
    const auto bonusPointColumn = c.data[0].indexOf("Bonuspunkte (Punkte)");

    // Throw if relevant columns not found
    if (registrationNrColumn == -1)
        throw BonusParseException(str::error::noIdColumn);
    if (bonusPointColumn == -1)
        throw BonusParseException(str::error::noBonusPointColumn);

    // Insert relevant data into lists
    for (qsizetype i = 1; i < nrRows; ++i) {

        // Registration number column sometimes contains an email address
        // instead of a registration number.
        // Because students are probably not listed under the same email address
        // in the dynexite/rwthonline file, don't add the email address as
        // registration number
        // Todo: Try to match them up via name?

        bool success;
        Points bonusPoints;
        c.data[i][registrationNrColumn].toInt(&success);
        try {
            bonusPoints = stringToPoints(c.data[i][bonusPointColumn]);
        } catch (...) {
            success = false;
        }

        if (success) {
            mRegistrationNr.emplace_back(c.data[i][registrationNrColumn]);
            mNrBonusPoints.emplace_back(bonusPoints);
        } else {
            mInvalidRegNrs.emplace_back(c.data[i][registrationNrColumn],
                                        c.data[i][bonusPointColumn]);
        }
    }

    if (mRegistrationNr.size() != mNrBonusPoints.size())
        throw BonusParseException(str::error::bonusArraySizeMismatch);
}
