#include "data_manager.hpp"

#include "file_io/d2r_parser.hpp"

#include <cassert>

DataManager &DataManager::getInstance() noexcept {
    // Guaranteed to be destroyed, instantiated on first use.
    static DataManager instance;

    return instance;
}

bool DataManager::loadNewExam(const DynexiteExam &exam,
                              const StudentData &students,
                              const UserGradeConfig &gradeConfig,
                              const UserPointConfig &pointConfig) noexcept {

    // Try to create a DAG object, if successful, move it into mNew,
    // else leave mNew untouched

    // Currently, creating a DAG cannot fail (except bad_alloc, which
    // we don't care about and results in immediate std::terminate)
    getInstance().mNew =
        std::make_unique<DAG>(exam, students, gradeConfig, pointConfig);

    return true;
}
void DataManager::makeNewCurrent() noexcept {
    getInstance().mCurrent = std::move(getInstance().mNew);
    // mNew is nullptr after this move

    emit getInstance().rawDataChanged();
    emit getInstance().gradesChanged(EnumStorage<bool>{true, true, true});
}

const DAG *DataManager::getCurrent() noexcept {
    return getInstance().mCurrent.get();
}

void DataManager::maxChange(const GradingType grading,
                            const GradeBoundary newMax,
                            DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return;

    switch (grading) {
    default: // fallthrough, "combined" not allowed
        assert(false);
        [[fallthrough]];
    case GradingType::Closed:
        configuredExam->mChangedMaxPointsClosed.setMax(newMax);
        configuredExam->updateMemberNodesIfNecessary<
            ChangedMaxPoints<GradingType::Closed>>();
        break;
    case GradingType::Open:
        configuredExam->mChangedMaxPointsOpen.setMax(newMax);
        configuredExam->updateMemberNodesIfNecessary<
            ChangedMaxPoints<GradingType::Open>>();
        break;
    }

    if (!alternateExam) {
        emit getInstance().gradesChanged(
            EnumStorage<bool>{grading == GradingType::Closed,
                              grading == GradingType::Open, true});
    }
}

bool DataManager::boundaryChange(const GradingType grading,
                                 const Grade boundaryGrade,
                                 const Points newBoundaryPoints,
                                 DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return false;

    if (std::isnan(newBoundaryPoints.p))
        return false;

    const GradeBoundary newBoundary(
        newBoundaryPoints,
        configuredExam->mUserPointConfig.mGradeBoundaryStepSize,
        configuredExam->mUserPointConfig.mGradeBoundaryRounding);

    return boundaryChange(grading, boundaryGrade, newBoundary, alternateExam);
}

bool DataManager::boundaryChange(const GradingType grading,
                                 const Grade boundaryGrade,
                                 const GradeBoundary newBoundary,
                                 DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return false;

    bool ret;

    switch (grading) {
    default: // fallthrough, "combined" not allowed
        assert(false);
        [[fallthrough]];
    case GradingType::Closed:
        ret = configuredExam->mClosedUserBoundaries.setBoundary(newBoundary,
                                                                boundaryGrade);
        configuredExam->updateMemberNodesIfNecessary<
            UserGradeBoundaries<GradingType::Closed>>();
        break;
    case GradingType::Open:
        ret = configuredExam->mOpenUserBoundaries.setBoundary(newBoundary,
                                                              boundaryGrade);
        configuredExam->updateMemberNodesIfNecessary<
            UserGradeBoundaries<GradingType::Open>>();
        break;
    }

    if (!alternateExam) {
        emit getInstance().gradesChanged(
            EnumStorage<bool>{grading == GradingType::Closed,
                              grading == GradingType::Open, true});
    }

    return ret;
}

void DataManager::autoBoundaryChange(const GradingType grading,
                                     const bool followDefault,
                                     DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return;

    switch (grading) {
    default: // fallthrough, "combined" not allowed
        assert(false);
        [[fallthrough]];
    case GradingType::Closed:
        configuredExam->mClosedUserBoundaries.setFollowDefaultAll(
            followDefault);
        configuredExam->updateMemberNodesIfNecessary<
            UserGradeBoundaries<GradingType::Closed>>();
        break;
    case GradingType::Open:
        configuredExam->mOpenUserBoundaries.setFollowDefaultAll(followDefault);
        configuredExam->updateMemberNodesIfNecessary<
            UserGradeBoundaries<GradingType::Open>>();
        break;
    }

    if (!alternateExam) {
        emit getInstance().gradesChanged(
            EnumStorage<bool>{grading == GradingType::Closed,
                              grading == GradingType::Open, true});
    }
}

void DataManager::gradeConfigChange(const bool requirePassAllParts,
                                    DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return;

    configuredExam->mUserGradeConfig.mRequirePassAllParts = requirePassAllParts;
    configuredExam->updateMemberNodesIfNecessary<UserGradeConfig>();

    if (!alternateExam) {
        // config changes (currently) only affect the combined grade
        emit getInstance().gradesChanged(
            getBoolStorage<3>(GradingType::Combined));
    }
}

void DataManager::pointConfigChange(const UserPointConfig newConfig,
                                    DAG *alternateExam) noexcept {
    auto *configuredExam =
        alternateExam ? alternateExam : getInstance().mCurrent.get();

    if (!configuredExam)
        return;

    configuredExam->mUserPointConfig = newConfig;
    configuredExam->updateMemberNodesIfNecessary<UserPointConfig>();

    if (!alternateExam) {
        emit getInstance().gradesChanged({true, true, true});
    }
}

void DataManager::applySettings(const D2rSettings &s) {

    const auto cur = getCurrent();

    if (!cur)
        return;

    if (s.mVersion.isEmpty())
        return;

    // userPointConfigStepSize already set

    // count zero
    {
        auto currentConfig = cur->mUserPointConfig;
        currentConfig.mCountZerothAttemptAsFirstAttempt =
            s.mUserPointConfigCountZero;
        pointConfigChange(currentConfig);
    }

    // require pass all
    gradeConfigChange(s.mUserGradeConfigRequirePassAll);

    // max boundary changes
    maxChange(GradingType::Closed, GradeBoundary{s.mChangedMaxPointsClosed});
    maxChange(GradingType::Open, GradeBoundary{s.mChangedMaxPointsOpen});

    // changedMaxPointsClosedFollowDefault isn't used at the moment (?)

    // autoBoundaryChange
    autoBoundaryChange(GradingType::Closed, s.mUserGradeBoundariesClosedAuto);
    autoBoundaryChange(GradingType::Open, s.mUserGradeBoundariesOpenAuto);

    // grade boundaries
    for (auto i = static_cast<int>(grades.size()) - 1; i >= 0; --i) {

        // make unsigned for array access
        const auto ui = static_cast<decltype(grades)::size_type>(i);

        boundaryChange(GradingType::Closed, grades[ui],
                       GradeBoundary{s.mUserGradeBoundariesClosed[ui]});
        boundaryChange(GradingType::Open, grades[ui],
                       GradeBoundary{s.mUserGradeBoundariesOpen[ui]});
    }
}
