#pragma once

/**
 * \file grade.hpp
 * \brief Declarations of the grade enum and related useful helper functions.
 */

#include "enums.hpp"

#include <QString>
#include <QtGlobal>

#include <array>
#include <cassert>

enum class Grade {
    grade1_0,
    grade1_3,
    grade1_7,
    grade2_0,
    grade2_3,
    grade2_7,
    grade3_0,
    grade3_3,
    grade3_7,
    grade4_0,
    grade5_0,
    NR_VALS
};

inline constexpr auto grades = std::to_array(
    {Grade::grade1_0, Grade::grade1_3, Grade::grade1_7, Grade::grade2_0,
     Grade::grade2_3, Grade::grade2_7, Grade::grade3_0, Grade::grade3_3,
     Grade::grade3_7, Grade::grade4_0, Grade::grade5_0});

inline constexpr auto gradeDoubles =
    std::to_array({1.0, 1.3, 1.7, 2.0, 2.3, 2.7, 3.0, 3.3, 3.7, 4.0, 5.0});

[[nodiscard]] inline auto toString(const Grade e,
                                   const QChar separator = '.') noexcept {
    switch (e) {
    case Grade::grade1_0:
        return QString("1") + separator + "0";
    case Grade::grade1_3:
        return QString("1") + separator + "3";
    case Grade::grade1_7:
        return QString("1") + separator + "7";
    case Grade::grade2_0:
        return QString("2") + separator + "0";
    case Grade::grade2_3:
        return QString("2") + separator + "3";
    case Grade::grade2_7:
        return QString("2") + separator + "7";
    case Grade::grade3_0:
        return QString("3") + separator + "0";
    case Grade::grade3_3:
        return QString("3") + separator + "3";
    case Grade::grade3_7:
        return QString("3") + separator + "7";
    case Grade::grade4_0:
        return QString("4") + separator + "0";
    case Grade::grade5_0:
        return QString("5") + separator + "0";
    case Grade::NR_VALS:
        return QString("NR_GRADES");
    default:
        return QString("unknown");
    }
}

constexpr Grade roundToNextBest(const double grade) noexcept {
    if (grade < 1.3)
        return Grade::grade1_0;
    if (grade < 1.7)
        return Grade::grade1_3;
    if (grade < 2.0)
        return Grade::grade1_7;
    if (grade < 2.3)
        return Grade::grade2_0;
    if (grade < 2.7)
        return Grade::grade2_3;
    if (grade < 3.0)
        return Grade::grade2_7;
    if (grade < 3.3)
        return Grade::grade3_0;
    if (grade < 3.7)
        return Grade::grade3_3;
    if (grade < 4.0)
        return Grade::grade3_7;
    if (grade == 4.0)
        return Grade::grade4_0;
    return Grade::grade5_0;
}

[[nodiscard]] constexpr double gradeToDouble(const Grade g) noexcept {

    if (static_cast<unsigned>(g) >= gradeDoubles.size()) {
        assert(false);
        return 5.0;
    }

    return gradeDoubles[static_cast<unsigned>(g)];
}

[[nodiscard]] constexpr bool
gradeBoundaryIsUserChangeable(const GradingType type, const Grade g) noexcept {
    using enum Grade;

    // 5.0 and invalid are not user changeable grade boundaries
    if (static_cast<unsigned>(g) > static_cast<unsigned>(grade4_0))
        return false;

    // The closed grade boundaries for 1.3, 2.3 and 3.3 are fixed
    return !(type == GradingType::Closed &&
             (g == grade1_3 || g == grade2_3 || g == grade3_3));
}

/**
 * \brief Computes the average grade of two grades with the given weights per
 * grade. If requirePassingBoth is set, the resulting grade can only be better
 * than 5.0 if neither of the input grades are 5.0.
 */
[[nodiscard]] constexpr Grade
getGradeAverage(const Grade g1, const double g1Weight, const Grade g2,
                const double g2Weight, const bool requirePassingBoth) noexcept {
    if (requirePassingBoth && (g1 == Grade::grade5_0 || g2 == Grade::grade5_0))
        return Grade::grade5_0;

    // Assert that the weights are sensible and add to 1.
    assert(g1Weight >= 0. && g1Weight <= 1.);
    assert(g2Weight >= 0. && g2Weight <= 1.);
    assert(qFuzzyCompare(g1Weight + g2Weight, 1.));

    const double preciseResult =
        (gradeToDouble(g1) * g1Weight) + (gradeToDouble(g2) * g2Weight);

    // The resulting double value, e.g. 1.67, will always be rounded towards the
    // next best grade, in that case 1.3.
    // Exception: when the grade is >4.0, it will always be rounded to 5.0.
    return roundToNextBest(preciseResult);
}
