#include "export_rwthonline.hpp"

#include "data/csv_dump.hpp"
#include "data/data_manager.hpp"
#include "file_io/csv_parser.hpp"
#include "file_io/file.hpp"
#include "util/enums.hpp"
#include "util/grade.hpp"
#include "util/points.hpp"

#include <QFileDialog>
#include <QString>

#include <cstdio>
#include <ranges>

qsizetype getColumn(const CsvLine &headerLine, const QString &headerText) {

    for (qsizetype i = 0; i < headerLine.size(); ++i) {

        if (QString::compare(headerLine[i], headerText, Qt::CaseInsensitive) ==
            0) {
            return i;
        }
    }
    return headerLine.size();
}

// todo: replace std::snprintf
[[nodiscard]] QString getRemark(const GradingType type, const qsizetype idx) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return "";

    // std::format would be way nicer, but it isn't widely supported yet
    char buf[16] = {};
    std::snprintf(buf, 16, "%.2f", cur->mExam.mPoints[type][idx].p);
    const auto grade = cur->getGradeList(type)[idx];
    return QString(toString(type)) + " questions: " + buf +
           " pts (grade: " + toString(grade, ',') + ")";
}

void exportKnownStudents(QWidget *parent) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, "Save RWTHonline data including grades", "",
        "Comma-separated values (CSV) file (*.csv);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {

        auto exportCsvDump = cur->mStudents.mRwthOnline;
        auto &exportCsv = exportCsvDump.data;

        if (exportCsv.empty())
            return;

        // Insert student data into csv
        {
            const auto gradeColumn = getColumn(exportCsv[0], "grade");
            const auto remarkColumn = getColumn(exportCsv[0], "remark");
            if (gradeColumn >= exportCsv[0].size() ||
                remarkColumn >= exportCsv[0].size())
                return;

            // Clear previous cell content
            for (auto &l : exportCsv | std::views::drop(1)) {

                l[gradeColumn] = "";
                l[remarkColumn] = "";
            }

            // Insert new content where we have data
            for (qsizetype i = 0; auto &l : exportCsv | std::views::drop(1)) {

                using enum GradingType;

                // Search for same student in dynexite data
                const auto idx = cur->mExam.mIdentifier.indexOf(
                    cur->mStudents.mRegistrationNr[i]);

                // Convert to index, or skip if not found
                if (idx == -1) {
                    ++i;
                    continue;
                }

                // Save result in new csv
                l[gradeColumn] =
                    toString(cur->getGradeList(Combined)[idx], ',');
                // Only print closed/open points if there were any to earn
                if (cur->mExam.mPointsMax[Closed] > Points()) {
                    l[remarkColumn] = getRemark(Closed, idx);
                }
                if (cur->mExam.mPointsMax[Closed] > Points() &&
                    cur->mExam.mPointsMax[Open] > Points()) {
                    l[remarkColumn] += " - ";
                }
                if (cur->mExam.mPointsMax[Open] > Points()) {
                    l[remarkColumn] += getRemark(Open, idx);
                }

                ++i;
            }

            // Mark students without data as missing
            for (auto &l : exportCsv | std::views::drop(1)) {

                if (l[gradeColumn].isEmpty()) {
                    l[gradeColumn] = "X";
                }
            }
        }

        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCsv(*stream, exportCsvDump);
    }
}

void exportUnknownStudents(QWidget *parent) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent,
        "Save RWTHonline data including grades of students not in the given "
        "RWTHonline file",
        "", "Comma-separated values (CSV) file (*.csv);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {
        CsvDump exportDump;
        auto &exportCsv = exportDump.data;

        // Header line
        exportCsv = {{"REGISTRATION_NUMBER", "GRADE", "REMARK"}};

        // Insert students not in the input RWTHonline file
        for (qsizetype i = 0; auto &id : cur->mExam.mIdentifier) {

            using enum GradingType;

            // Search for same student in RWTHonline
            const auto pos = cur->mStudents.mRegistrationNr.indexOf(id);

            // Skip if found
            if (pos != -1) {
                ++i;
                continue;
            }

            // Save result in new line
            exportCsv.emplace_back(3);
            exportCsv.back()[0] = id;
            exportCsv.back()[1] = toString(cur->getGradeList(Combined)[i], ',');
            // Only print closed/open points if there were any to earn
            if (cur->mExam.mPointsMax[Closed] > Points()) {
                exportCsv.back()[2] = getRemark(Closed, i);
            }
            if (cur->mExam.mPointsMax[Closed] > Points() &&
                cur->mExam.mPointsMax[Open] > Points()) {
                exportCsv.back()[2] += " - ";
            }
            if (cur->mExam.mPointsMax[Open] > Points()) {
                exportCsv.back()[2] += getRemark(Open, i);
            }

            ++i;
        }

        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCsv(*stream, exportDump);
    }
}

int getNrUnknownStudents() noexcept {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return 0;

    int count = 0;

    for (const auto &id : cur->mExam.mIdentifier | std::views::drop(1)) {

        const auto pos = cur->mStudents.mRegistrationNr.indexOf(id);

        // Not found?
        if (pos == -1) {
            ++count;
        }
    }

    return count;
}
