#include "students_tab.hpp"

#include "data/data_manager.hpp"
#include "util/enums.hpp"
#include "util/grade.hpp"
#include "util/points.hpp"
#include "util/strings.hpp"

#include <QBrush>
#include <QLabel>
#include <QList>
#include <QMessageBox>
#include <QResizeEvent>
#include <QString>
#include <QTableWidgetItem>
#include <QTimer>
#include <QVBoxLayout>

#include <array>

static constexpr std::array tableHeaders = {"Student",      "Closed points",
                                            "Closed Grade", "Open points",
                                            "Open Grade",   "Combined Grade"};

// The background color of a grade cell changes depending on the grade
[[nodiscard]] QBrush getGradeBackgroundColor(const Grade g) noexcept {
    switch (g) {
    case Grade::grade1_0:
        return {Qt::green};
    case Grade::grade5_0:
        return {Qt::red};
    default:
        return {Qt::darkGreen};
    }
}

StudentsTab::StudentsTab(QWidget *parent) : QWidget(parent) {

    const auto layout = new QVBoxLayout(this);

    const auto legendLabel = new QLabel(str::info::studentsTableTabLegend);

    layout->addWidget(legendLabel);
    layout->addWidget(&mTable);

    connect(&DataManager::getInstance(), &DataManager::rawDataChanged, this,
            &StudentsTab::pointsChanged);
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &StudentsTab::gradesChanged);
}

void StudentsTab::pointsChanged() {
    const auto cur = DataManager::getCurrent();

    // Clear content and headers
    mTable.clear();

    if (!cur) {
        return;
    }

    // Column 0: Student IDs
    // Columns 1-2: Closed points+grades (if hasClosed)
    // Columns 3-4: Open points+grades (if hasOpen)
    // Column 5: Combined grade (if hasClosed && hasOpen)
    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};
    mTable.setColumnCount(1 + 2 * static_cast<int>(hasClosed) +
                          2 * static_cast<int>(hasOpen) +
                          static_cast<int>(hasClosed && hasOpen));

    // Set horizontal headers
    {
        int columnIdx = 0;

        mTable.setHorizontalHeaderItem(columnIdx++,
                                       new QTableWidgetItem(tableHeaders[0]));

        if (hasClosed) {

            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[1]));
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[2]));
        }
        if (hasOpen) {

            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[3]));
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[4]));
        }

        if (hasClosed && hasOpen) {
            mTable.setHorizontalHeaderItem(
                columnIdx++, new QTableWidgetItem(tableHeaders[5]));
        }
    }

    // Sorting has to be disabled while inserting items
    mTable.setSortingEnabled(false);

    const auto &ids = cur->mExam.mIdentifier;
    const auto &closedPoints = cur->mExam.mPoints[GradingType::Closed];
    const auto &openPoints = cur->mExam.mPoints[GradingType::Open];

    mTable.setRowCount(static_cast<int>(ids.size()));

    // Set student IDs and point numbers
    for (qsizetype row = 0; row < ids.size(); ++row) {

        int columnIdx = 0;

        // Student ID
        {
            const auto idItem = new QTableWidgetItem;
            idItem->setData(Qt::DisplayRole, ids[row]);
            idItem->setFlags(idItem->flags() ^ Qt::ItemIsEditable);
            mTable.setItem(static_cast<int>(row), columnIdx++, idItem);
        }

        // Closed
        if (hasClosed) {
            // Points
            {
                const auto closedPointsItem = new QTableWidgetItem;
                closedPointsItem->setData(Qt::DisplayRole, closedPoints[row].p);
                closedPointsItem->setFlags(closedPointsItem->flags() ^
                                           Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               closedPointsItem);
            }

            // Grade
            {
                const auto closedGradeItem = new QTableWidgetItem;
                closedGradeItem->setFlags(closedGradeItem->flags() ^
                                          Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               closedGradeItem);
            }
        }

        // Open
        if (hasOpen) {
            // Points
            {
                const auto openPointsItem = new QTableWidgetItem;
                openPointsItem->setData(Qt::DisplayRole, openPoints[row].p);
                openPointsItem->setFlags(openPointsItem->flags() ^
                                         Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               openPointsItem);
            }

            // Grade
            {
                const auto openGradeItem = new QTableWidgetItem;
                openGradeItem->setFlags(openGradeItem->flags() ^
                                        Qt::ItemIsEditable);
                mTable.setItem(static_cast<int>(row), columnIdx++,
                               openGradeItem);
            }
        }

        // Combined grade
        if (hasClosed && hasOpen) {
            const auto combinedGradeItem = new QTableWidgetItem;
            combinedGradeItem->setFlags(combinedGradeItem->flags() ^
                                        Qt::ItemIsEditable);
            mTable.setItem(static_cast<int>(row), columnIdx++,
                           combinedGradeItem);
        }
    }
    mTable.setSortingEnabled(true);

    // Since the font size might change, wait until that is done and then fix
    // column width and row height
    QTimer::singleShot(1000, this, [this] {
        mTable.resizeColumnsToContents();
        mTable.resizeRowsToContents();
    });
}

void StudentsTab::gradesChanged(const EnumStorage<bool> &changedTypes) {

    using enum GradingType;

    const auto cur = DataManager::getCurrent();

    if (!cur) {
        return;
    }

    mTable.setSortingEnabled(false);

    // Whether the given GradingType is present in this exam
    EnumStorage<bool> hasType;
    hasType[Closed] = cur->mExam.mPointsMax[Closed] > Points{0.};
    hasType[Open] = cur->mExam.mPointsMax[Open] > Points{0.};
    hasType[Combined] = hasType[Closed] && hasType[Open];

    const auto nrRows = mTable.rowCount();

    // Data structures for error handling
    QList<QString> notFoundIds;

    // Find corresponding grade to the student id in the row
    for (int i = 0; i < nrRows; ++i) {
        auto id = mTable.item(i, 0)->text();

        const auto idx = cur->mExam.mIdentifier.indexOf(id);

        // Found?
        if (idx == -1) {
            notFoundIds.push_back(id);
            continue;
        }

        EnumStorage<int> columnNrs{};
        columnNrs[Closed] = 2;
        columnNrs[Open] = hasType[Closed] ? 4 : 2;
        columnNrs[Combined] = 5;

        // Set new grade strings and new background color
        // Use strings instead of doubles for a fixed number of decimal places
        for (const auto type : {Closed, Open, Combined}) {
            if (hasType[type] && changedTypes[type]) {
                const auto item = mTable.item(i, columnNrs[type]);
                const auto grade = cur->getGradeList(type)[idx];
                item->setText(toString(grade));
                item->setBackground(getGradeBackgroundColor(grade));
            }
        }
    }

    // Display warnings if some occurred
    if (!notFoundIds.empty()) {
        QString informativeText = "";
        for (const auto &id : notFoundIds) {
            informativeText += id + "\n";
        }

        QMessageBox msgBox(this);
        msgBox.setTextFormat(Qt::TextFormat::MarkdownText);
        msgBox.setWindowTitle("Students-Tab warning.");
        msgBox.setText("**Failed to look up grades for the following\nstudent "
                       "ids in the students tab:**\n");
        msgBox.setInformativeText(informativeText);
        msgBox.setStandardButtons(QMessageBox::Ok);
        msgBox.exec();
    }

    mTable.setSortingEnabled(true);
}
